"""Bible classes."""

__version__ = '1.1.0'

import copy
import json
import os
from dataclasses import dataclass
from typing import Final, Union

import requests
from bs4 import BeautifulSoup, NavigableString
from bs4.element import Tag

from .data import BOOKS, CHAPTERS
from .utils import read_file, write_file

# Replace [EMAIL] with your own - let's be nice to The Lockman Foundation :)
USERAGENT: Final = "Python LSB Reader - [EMAIL]"

CLASS_TO_ATTR_MAPPING: Final = {
    'single-quote': 'quote#{}#1',
    'double-quote': 'quote#{}#2',
    'block-quote': 'indent#{}#1',
    'indented-block-quote': 'indent#{}#2',
    'poetry': 'poetry#{}#1',
}


@dataclass
class BibleMarker:
    """A specific verse in the Bible."""

    book: str
    chapter: int
    verse: int


@dataclass
class BibleRange:
    """A range of verse."""

    start: BibleMarker
    end: BibleMarker


def expand_book(book: str):
    """Attempt to get a Bible book based on the starting characters.

    For example, if you passed MAT it would return matthew. However, if you
    passed m, since that matches micah, malachi, matthew, and mark, we cannot
    determine the correct book, so an empty string will be returned. If the
    string passed doesn't match the start of ANY book, None is returned.
    """
    book = book.lower()
    if book in BOOKS:
        return book
    candidates = [_b for _b in BOOKS if _b.startswith(book)]
    if not candidates:
        return None
    if len(candidates) > 1:
        return ''
    return candidates[0]


class BibleGetter():
    """Manages a collection of Bible verses, downloading them when needed."""

    URL: Final = "https://read.lsbible.org/"

    def __init__(self, file):
        """Initialize Bible data."""
        self.file = file
        self.data = (json.loads(read_file(file)) if os.path.exists(file) else {
            book: {'chapters': CHAPTERS[idx]}
            for idx, book in enumerate(BOOKS)})
        self.changed = False

    def __get_chapter(self, book: str, chapter: int) -> dict:
        """Download a chapter of the Bible."""
        # print(f"Caching {book} {chapter}...")
        _r = requests.get(
            self.URL,
            params={'q': f'{book}{chapter}'},
            headers={'user-agent': USERAGENT},
            timeout=30)
        _bs = BeautifulSoup(_r.text, 'lxml')
        verses = _bs.find_all('span', class_='verse')
        self.data[book][str(chapter)] = {
            str(int(verse.attrs['data-key'][-3:])):
                BibleGetter.__jsonify_verse(
                    verse,
                    _bs,
                    int(verse.attrs['data-key'][-3:]))
            for verse in verses
        }
        self.data[book][str(chapter)]['verses'] = max({
            int(span.attrs['data-key'][-3:])
            for span in verses})
        self.changed = True
        if chapter == 1 or book == 'psalm':
            self.data[
                book][str(chapter)]['1']['attributes'].append('paragraph')
        return self.data[book][str(chapter)]

    @classmethod
    def __jsonify_verse(
        cls,
        verse: Tag,
        _bs: BeautifulSoup,
        verse_num: int
    ) -> dict:
        """Create a database entry for an HTML verse tag."""
        lines = [
            _c
            for _c in verse.contents
            if (_c.strip() if isinstance(_c, NavigableString) else (
                _c.get('class', '')
                and {
                    'prose',
                    'block-quote',
                    'poetry',
                    'indented-block-quote'
                }.intersection(set(_c['class']))))]
        attrs = []
        for idx, line in enumerate(lines):
            # Turn non-tagged strings into paragraphs
            if isinstance(line, str):
                _p = _bs.new_tag('p', attrs={'class': ''})
                _p.string = line
                line = _p
                lines[idx] = line
            for class_name, attr_template in CLASS_TO_ATTR_MAPPING.items():
                if class_name in line['class']:
                    attrs.append(attr_template.format(idx))
            if line.text.strip().endswith(' Higgaion Selah.'):
                attrs.append(f'higgaion-selah#{idx}#1')
                string = list(line.strings)[-1]
                string.replace_with(string.strip()[:-16])
            if line.text.strip().endswith(' Selah.'):
                attrs.append(f'selah#{idx}#1')
                string = list(line.strings)[-1]
                string.replace_with(string.strip()[:-7])
        # I wonder if these can be replaced with just find instead of find_all
        if verse.find_all('span', class_='start-pericope', recursive=False):
            attrs.append('paragraph')
        if verse.find_all('br'):
            attrs.append('break')
        search = verse.find_all('h3', class_='hebrew-letter')
        if search:
            attrs.append('acrostic')
            attrs.append(f'acrostic#{search[0].text}')
        heading = ''
        search = verse.find_all('h3', class_='subhead')
        if search:
            heading = search[0].text
            if verse_num == 1:
                attrs.append('paragraph')
        subheading = ''
        search = verse.find_all('div', class_='included-subhead')
        if search:
            subheading = search[0].text
        return {
            'attributes': attrs,
            'heading': heading,
            'subheading': subheading,
            'text': [BibleGetter.__stringify_line(_v).text for _v in lines]
        }

    def __list(
        self,
        selection: BibleRange,
        redownload: bool
    ) -> list[BibleMarker]:
        """Generate a list of each verse referenced in a BibleRange."""
        ret = []
        # Get book range (since these are non-numeric)
        books = list(self.data.keys())
        for book in books[
            books.index(selection.start.book):
            books.index(selection.end.book) + 1
        ]:
            for chapter in range(
                1 if selection.start.book != book else selection.start.chapter,
                (
                    self.data[book]['chapters']
                    if selection.end.book != book
                    else selection.end.chapter
                ) + 1
            ):
                if redownload or str(chapter) not in self.data[book]:
                    self.__get_chapter(book, chapter)
                ret.extend([
                    BibleMarker(book, chapter, verse)
                    for verse in range(
                        (
                            1
                            if selection.start.book != book
                            or selection.start.chapter != chapter
                            else selection.start.verse
                        ),
                        (
                            self.data[book][str(chapter)]['verses']
                            if selection.end.book != book
                            or selection.end.chapter != chapter
                            else selection.end.verse
                        ) + 1
                    )])
        # print(f"Returning: {ret}")
        return ret

    @classmethod
    def __stringify_line(cls, line: Tag) -> str:
        """Stringifies certain elements of an HTML verse tag."""
        for caps in line.find_all('span', class_='small-caps'):
            for _s in list(caps.strings):
                _s.replace_with(_s.upper())
        for red in line.find_all('span', class_='red-letter'):
            for _s in list(red.strings):
                _s.replace_with(f'\033[31m{_s.text}\033[39m')
        for italic in line.select('i'):
            italic.replace_with(f'\033[3m{italic.text}\033[23m')
        return line

    def find_attribute_backwards(
        self,
        start: BibleMarker,
        attr: str,
        inclusive: bool = False,
        redownload: bool = False
    ) -> BibleMarker:
        """Search for the first verse with attr, towards Genesis 1."""
        prev_ = start
        next_ = copy.copy(start)
        if redownload:
            self.data[next_.book].pop(str(next_.chapter))
        while True:
            if str(next_.chapter) not in self.data[next_.book]:
                self.__get_chapter(next_.book, next_.chapter)
            if (
                attr in self.data
                    [next_.book][str(next_.chapter)][str(next_.verse)]
                    ['attributes']
            ):
                break
            if next_.verse == 1:
                if next_.chapter == 1:
                    if next_.book == 'genesis':
                        break
                    prev_ = copy.copy(next_)
                    idx = BOOKS.index(next_.book)
                    next_.book = BOOKS[idx - 1]
                    next_.chapter = CHAPTERS[idx - 1]
                else:
                    prev_ = copy.copy(next_)
                    next_.chapter -= 1
                if (
                    redownload
                    or str(next_.chapter) not in self.data[next_.book]
                ):
                    self.__get_chapter(next_.book, next_.chapter)
                next_.verse = self.data[
                    next_.book][str(next_.chapter)]['verses']
            else:
                prev_ = copy.copy(next_)
                next_.verse -= 1
        return next_ if inclusive else prev_

    def find_attribute_forwards(
        self,
        start: BibleMarker,
        attr: str,
        inclusive: bool = False,
        redownload: bool = False
    ) -> BibleMarker:
        """Search for the first verse with attr, towards Revelation 22."""
        prev_ = start
        next_ = copy.copy(start)
        if redownload:
            self.data[next_.book].pop(str(next_.chapter))
        while True:
            if str(next_.chapter) not in self.data[next_.book]:
                self.__get_chapter(next_.book, next_.chapter)
            if (
                next_ != start
                and attr in self.data
                    [next_.book][str(next_.chapter)][str(next_.verse)]
                    ['attributes']
            ):
                break
            if (
                next_.verse == self.data[
                    next_.book][str(next_.chapter)]['verses']
            ):
                if next_.chapter == self.data[next_.book]['chapters']:
                    if next_.book == 'revelation':
                        break
                    prev_ = copy.copy(next_)
                    idx = BOOKS.index(next_.book)
                    next_.book = BOOKS[idx + 1]
                    next_.chapter = 0
                else:
                    prev_ = copy.copy(next_)
                next_.chapter += 1
                next_.verse = 0
            else:
                prev_ = copy.copy(next_)
            next_.verse += 1
        return next_ if inclusive else prev_

    def get(
        self,
        selection: Union[BibleMarker, BibleRange],
        redownload: bool
    ) -> dict:
        """Retrieve a specified verse, or range of verses."""
        if not self.valid(selection):
            return {}
        if isinstance(selection, BibleMarker):
            selection = BibleRange(selection, selection)
        ret = {}
        for marker in self.__list(selection, redownload):
            if marker.book not in ret:
                ret[marker.book] = {}
            if marker.chapter not in ret[marker.book]:
                ret[marker.book][marker.chapter] = {}
            ret[marker.book][marker.chapter][marker.verse] = (
                self.data[marker.book][str(marker.chapter)][str(marker.verse)])
        return ret

    def get_chapters(self, book: str) -> int:
        """Get the chapter count for a particular book of the Bible."""
        return self.data[book]['chapters'] if book in self.data else 0

    def get_verses(self, book: str, chapter: int) -> int:
        """Get the verse count for a particular chapter of the Bible."""
        if (
            book not in self.data
            or chapter < 1
            or chapter > self.data[book]['chapters']
        ):
            return 0
        if str(chapter) not in self.data[book]:
            self.__get_chapter(book, chapter)
        return self.data[book][str(chapter)]['verses']

    def highlight(self, book: str, chapter: int, verse: int, color: int):
        """Highlight a verse.

        If color is 0, the highlight is removed.
        """
        attrs = self.data[book][str(chapter)][str(verse)]['attributes']
        attrs = [
            _a
            for _a in attrs
            if not _a.startswith('highlight#')]
        if color != 0:
            attrs.append(f'highlight#{color}')
        self.changed = True
        self.data[book][str(chapter)][str(verse)]['attributes'] = attrs

    def save(self):
        """Write back data if changed."""
        if self.changed:
            write_file(self.file, json.dumps(self.data))
            self.changed = False

    def valid(self, selection: Union[BibleMarker, BibleRange]) -> bool:
        """Check if a Scripture reference is valid.

        If a chapter is valid, and is not downloaded yet, it gets downloaded
        before this function returns.
        """
        if isinstance(selection, BibleMarker):
            return (
                selection.book in self.data
                and selection.chapter > 0
                and selection.chapter <= self.data[selection.book]['chapters']
                and selection.verse > 0
                and (
                    str(selection.chapter) in self.data[selection.book]
                    or self.__get_chapter(
                        selection.book,
                        selection.chapter)
                )
                and selection.verse <= self.data[
                    selection.book][str(selection.chapter)]['verses'])
        # Is range
        return self.valid(selection.start) and self.valid(selection.end)



"""Argument parsing for PyLSB."""

__version__ = '1.0.0'

import argparse
import random
import re
import sys
from typing import Union

from .bible import BibleGetter, BibleMarker, BibleRange, expand_book
from .data import BOOKS

VERSE = r"(\d?)\s*([a-zA-Z]+)\s*(\d+)(?::(\d+))?"
VERSE_OPT = r"(\d??)\s*([a-zA-Z]+)?\s*(\d+)(?::(\d+))?"
BIBLE_REGEX = re.compile(rf"^\s*{VERSE}\s*(?:-\s*{VERSE_OPT}\s*)?$")


def parse_scripture(arg: str, bible: BibleGetter) -> BibleRange:
    """Parse a verse string into a BibleRange.

    Returns None if it fails to parse.
    """
    # Read a verse or range of verses
    match = BIBLE_REGEX.match(arg)
    if not match:
        raise ValueError(f'Unable to parse {arg} as a Bible verse or range.')
    start = BibleMarker(
        f"{match.group(1) or ''}{match.group(2)}",
        int(match.group(3)),
        -1 if not match.group(4) else int(match.group(4)))
    end = BibleMarker(
        f"{match.group(5) or ''}"
        f"{match.group(6) or ''}",
        -1 if not match.group(7) else int(match.group(7)),
        -1 if not match.group(8) else int(match.group(8)))
    # Expand shortened name
    _sb = expand_book(start.book)
    if _sb is None:
        raise ValueError(
            f'{start.book} '
            'does not match the start of any book of the Bible.')
    if not _sb:
        raise ValueError(
            f'{start.book} matches more than one book of the Bible.')
    start.book = _sb
    # Make sure chapter is valid
    if start.chapter > bible.get_chapters(start.book):
        raise ValueError(
            f"{start.book.title()} doesn't have that many chapters!")
    return BibleRange(start, end)


def get_scripture(
    arg: str,
    bible: BibleGetter
) -> Union[BibleMarker, BibleRange]:
    """Get the appropriate BibleMarker or BibleRange for a string.

    Returns None if it fails to parse, or if invalid values are given.
    """
    # Custom arguments
    if arg == 'random':
        book = random.choice(BOOKS)
        chapter = random.choice(range(0, bible.get_chapters(book))) + 1
        return BibleMarker(
            book,
            chapter,
            random.choice(range(0, bible.get_verses(book, chapter))) + 1)
    # Standard argument
    range_ = parse_scripture(arg, bible)
    # No range explicitly given
    if range_.end.chapter == -1:
        # Single verse
        if range_.start.verse != -1:
            return range_.start
        # No starting verse specified, so give the whole chapter
        range_.start.verse = 1
        range_.end = BibleMarker(
            range_.start.book,
            range_.start.chapter,
            bible.get_verses(range_.start.book, range_.start.chapter))
    # Explicit range
    else:
        # Different book
        if range_.end.book:
            # Expand shortened name
            _eb = expand_book(range_.end.book)
            if _eb is None:
                raise argparse.ArgumentTypeError(
                    f'{range_.end.book} '
                    'does not match the start of any book of the Bible.')
            if not _eb:
                raise argparse.ArgumentTypeError(
                    f'{range_.end.book} '
                    'matches more than one book of the Bible.')
            range_.end.book = _eb
        # No book specified (same book)
        else:
            # Make sure chapter isn't actually a verse
            if range_.end.verse == -1 and range_.start.verse != -1:
                range_.end.verse = range_.end.chapter
                range_.end.chapter = range_.start.chapter
            range_.end.book = range_.start.book
        # Make sure chapter is valid
        if range_.end.chapter > bible.get_chapters(range_.end.book):
            raise argparse.ArgumentTypeError(
                f'{range_.end.book.title()} '
                "doesn't have that many chapters!")
        # Fill in missing data where applicable
        if range_.start.verse == -1:
            range_.start.verse = 1
        if range_.end.verse == -1:
            range_.end.verse = bible.get_verses(
                range_.end.book,
                range_.end.chapter)
    return range_


def get_highlight(val: str) -> int:
    """Parse a string as either a 1-3 digit integer, or a 6 digit hex value."""
    val = val.strip()
    if val.startswith('-'):
        raise argparse.ArgumentTypeError('color must be a positive value')
    if len(val) < 4:
        return int(val)
    if len(val[2:] if val.startswith('0x') else val) != 6:
        raise argparse.ArgumentTypeError('color should be 1, 2, 3 or 6 digits')
    return -int(val, 16)


def args(bible: BibleGetter) -> list[Union[BibleMarker, BibleRange]]:
    """Parse sys.argv into Scripture and options."""
    parser = argparse.ArgumentParser(
        usage='%(prog)s [options] scripture [scripture ...]',
        description=(
            'Downloads, displays, and caches verses '
            'from the Legacy Standard Bible'),
        epilog=(
            'The scripture references and options '
            'may be passed in any order.'))
    parser.add_argument(
        'scripture',
        action='store',
        help=(
            'chapter, verse, range of chapters or verses, '
            'or "random" for a random verse'),
        nargs='+',
        type=lambda _s: get_scripture(_s.lower(), bible))
    parser.add_argument(
        '-p',
        '--paragraph',
        action='store_true',
        help='expand a verse or range to include the whole paragraph')
    parser.add_argument(
        '-r',
        '--redownload',
        action='store_true',
        help='force a cache update even if the verses are already downloaded')
    parser.add_argument(
        '-n',
        '--nored',
        action='store_true',
        help='disable red text quotes')
    parser.add_argument(
        '-8',
        '--80columns',
        action='store_true',
        help='force width to 80 columns (ignore terminal settings)',
        dest='columns80')
    parser.add_argument(
        '-w',
        '--web',
        action='store_true',
        help='format text like the web version instead of the print version')
    parser.add_argument(
        '-b',
        '--browser',
        action='store_true',
        help='open scripture in web browser')
    parser.add_argument(
        '-f',
        '--preface',
        action='store_true',
        help='display preface before scripture')
    parser.add_argument(
        '-m',
        '--highlight',
        action='store',
        help=(
            'highlight all selected verses with specified color num '
            '(0 to remove)'),
        metavar='color',
        type=get_highlight)
    parser.add_argument(
        '-2',
        '--split',
        action='store_true',
        help='print verses in 2 columns, one screen at a time')
    return parser.parse_args(sys.argv[1:])



#!/usr/bin/python3
"""General functions shared between my scripts."""

__version__ = '1.0.0'


def read_file(filename: str):
    """Read contents of file, close, and then return contents."""
    with open(filename, "r", encoding="UTF-8") as _f:
        return _f.read()


def write_file(filename: str, data: str):
    """Write data to file (does not append), and then close."""
    with open(filename, "w", encoding="UTF-8") as _f:
        _f.write(data)
